import request from "supertest";
import { server } from "@/server";
import { mockConfig } from "@/../test_setup/mockConfig";
import { Email } from "@/email/email";

const config = {
    send_email: false,
};

const resetConfig = () => {
    config.send_email = false;
};

let user_email = "test@test.de";
vi.mock("../../email/send", async (importActual) => {
    const actual = await importActual<typeof import("../../email/send")>();
    return {
        sendEmail: async (email: Email) => {
            if (config.send_email) {
                return await actual.sendEmail(email);
            }
            return;
        },
    };
});

let user_not_found = false;
vi.mock("@snip/database", () => {
    return {
        default: {
            user: {
                getByEmail: async (email: string) => {
                    if (email.includes("notfound") || user_not_found) {
                        throw new Error("User not found");
                    }
                    if (email.includes("error")) {
                        throw new Error("Test error");
                    }
                    return {
                        email,
                    };
                },
                getById: async (id: number) => {
                    if (id === 1) {
                        return {
                            email: user_email,
                        };
                    } else {
                        throw new Error("User not found");
                    }
                },
                getPasswordHash: async (email: string) => {
                    return "test";
                },
                loginById: async (id: number, password: string) => {
                    if (id === 1 && password === "valid") {
                        return true;
                    }
                    if (password === "error") {
                        throw new Error("Test error");
                    }
                    return false;
                },
            },
        },
    };
});

describe("account router", () => {
    let consoleSpy;
    beforeAll(() => {
        consoleSpy = vi.spyOn(console, "error").mockImplementation(() => {});
    });

    beforeEach(() => {
        resetConfig();
    });

    afterAll(() => {
        server.close();
    });

    describe("POST /email/account/passwordReset/:email", () => {
        it("return 400 on invalid email", async () => {
            const res = await request(server)
                .post("/email/account/passwordReset/invalidemail")
                .set("X-Forwarded-For", "192.168.2.1")
                .expect(400)
                .expect("Content-Type", /json/);
            expect(res.body).toEqual({
                error: "Invalid email address",
            });
        });

        it("return 200 on valid email and send email", async () => {
            const res = await request(server)
                .post("/email/account/passwordReset/sebastian@test.de")
                .set("X-Forwarded-For", "192.168.2.2")
                .expect(200);
            expect(res.body).toEqual({
                success: true,
            });
        });

        it("should return 400 user not found", async () => {
            const res = await request(server)
                .post("/email/account/passwordReset/sebastian@notfound.de")
                .set("X-Forwarded-For", "192.168.2.3")
                .expect(400)
                .expect("Content-Type", /json/);
            expect(res.body).toEqual({
                error: "User not found",
            });
        });

        it("should return 429 on rate limit", async () => {
            let res: any;
            for (let i = 0; i < 10; i++) {
                res = await request(server)
                    .post("/email/account/passwordReset/sebastian@notfound.de")
                    .set("X-Forwarded-For", "192.168.2.5")
                    .expect("Content-Type", /json/);
            }
            expect(res.body).toEqual({
                error: "Too Many Requests",
                details: "Please wait 5s between each request",
            });
        });
    });

    describe("POST /email/account/verifyEmail/:email", () => {
        it("return 200 on valid email and send email", async () => {
            const res = await request(server)
                .post("/email/account/verifyEmail")
                .set("X-Forwarded-For", "192.168.3.2")
                .expect(200);
            expect(res.body).toEqual({
                success: true,
            });
        });

        it("should return 400 user not found", async () => {
            mockConfig.session!.user!.id = 2;

            const res = await request(server)
                .post("/email/account/verifyEmail")
                .set("X-Forwarded-For", "192.168.3.3")
                .expect(400)
                .expect("Content-Type", /json/);
            expect(res.body).toEqual({
                error: "User not found",
            });

            mockConfig.session!.user!.id = 1;
        });

        it("should return 429 on rate limit", async () => {
            let res: any;
            for (let i = 0; i < 10; i++) {
                res = await request(server)
                    .post("/email/account/verifyEmail")
                    .set("X-Forwarded-For", "192.168.3.5")
                    .expect("Content-Type", /json/);
            }
            expect(res.body).toEqual({
                error: "Too Many Requests",
                details: "Please wait 5s between each request",
            });
        });
    });

    /** Change Email
     *
     */
    describe("POST /email/account/changeEmail", () => {
        it("return 400 on invalid body", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.1")
                .expect(400)
                .expect("Content-Type", /json/);
            expect(res.body).toEqual({
                error: "Invalid request body",
                details: "Valid Email and password required",
            });
        });

        it("return 400 on invalid email", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.2")
                .send({
                    email: "invalidemail",
                    password: "valid",
                })
                .expect(400)
                .expect("Content-Type", /json/);
            expect(res.body).toEqual({
                error: "Invalid request body",
                details: "Valid Email and password required",
            });
        });

        it("return 400 on invalid password", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.3")
                .send({
                    email: "test@test.de",
                    password: "does not match set pw",
                })
                .expect(400)
                .expect("Content-Type", /json/);

            expect(res.body).toEqual({
                error: "Invalid password",
            });
        });
        it("return 400 if email is already registered", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.4")
                .send({
                    email: "test@test.de",
                    password: "valid",
                });

            expect(res.body).toEqual({
                error: "Email already registered on other account!",
            });
        });

        it("handle error in loginById", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.5")
                .send({
                    email: "test@test.de",
                    password: "error",
                });
            expect(res.body).toEqual({
                error: "Invalid password",
            });
        });

        it("handle error in getByEmail", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.6")
                .send({
                    email: "error@test.de",
                    password: "valid",
                });
            expect(res.body).toEqual({
                error: "Internal server error",
                details: "Test error",
            });
        });

        it("should register a new email", async () => {
            const res = await request(server)
                .post("/email/account/changeEmail")
                .set("X-Forwarded-For", "192.168.4.7")
                .send({
                    email: "notfound@test.de",
                    password: "valid",
                });
            expect(res.body).toEqual({
                success: true,
            });
        });
    });

    /** Uncomment to send emails as test!
     *
    describe("SEND EMAILS", () => {
        const addr = "youremailhere@uni-goettingen.de";
        it("verifyEmail", async () => {
            send_email = true;
            user_email = addr;
            await request(server)
            .post("/email/account/verifyEmail/")
            .set("X-Forwarded-For", "192.168.99.1")
            .expect(200);
            user_email = "test@test.de";
        });
        it("passwordReset", async () => {
            send_email = true;
            await request(server)
            .post("/email/account/passwordReset/" + addr)
            .set("X-Forwarded-For", "192.168.99.2")
            .expect(200);
        });
        it("changeEmail", async () => {
            send_email = true;
            user_not_found = true;
            const res = await request(server)
            .post("/email/account/changeEmail")
            .set("X-Forwarded-For", "192.168.99.3")
            .send({
                email: addr,
                password: "valid",
            })
            .expect(200);
            
            expect(res.body).toEqual({
                success: true,
            });
            
            send_email = false;
            user_not_found = false;
        });
    });
    */
});
