"use client";

import { useDate } from "lib/hooks/useDate";
import useUser from "lib/hooks/useUser";
import { useState } from "react";
import { Button, Modal } from "react-bootstrap";
import { useFormState } from "react-dom";

import { changePasswordAction, ChangePasswordState } from "./changePwAction";
import { UserProfileProvider, useUserProfile } from "./context";

import { FormError } from "components/form/alert";
import { SubmitButton } from "components/form/buttons";
import { EmailInput, PasswordInput } from "components/form/inputs/text";
import { Item } from "components/form/items";
import { validateEmail, validatePW } from "components/form/validation";

import styles from "./profileN.module.scss";

export default function ProfileOverviewPage() {
    return (
        <UserProfileProvider>
            <h2>General profile settings</h2>
            <Item
                title="Email address"
                description="Changing your email address requires an additional password validation and email confirmation on the new email address."
            >
                <ChangeEmail />
            </Item>
            <Item
                title="Change password"
                description="Changing your password requires to use your current password or verification via email."
            >
                <ChangePassword />
            </Item>
            <Item
                title="Last login"
                description="The last time you logged in. This might be helpful to verify if someone else accessed your account."
            >
                <LastLogin />
            </Item>
            <Item
                title="User data"
                description="You can download all your user data in a JSON format. This includes all your personal data, settings, and groups. This does not include any book data."
            >
                <DownloadUserData />
            </Item>
        </UserProfileProvider>
    );
}

function DownloadUserData() {
    const { user, config } = useUserProfile();

    function downloadUserData() {
        saveTemplateAsFile(`${user?.email}.json`, {
            user,
            config,
        });
    }

    return (
        <Button variant="outline-dark" onClick={downloadUserData}>
            Download user data (JSON)
        </Button>
    );
}

const saveTemplateAsFile = (filename: string, dataObjToWrite: unknown) => {
    const blob = new Blob([JSON.stringify(dataObjToWrite)], {
        type: "text/json",
    });
    const link = document.createElement("a");

    link.download = filename;
    link.href = window.URL.createObjectURL(blob);
    link.dataset.downloadurl = ["text/json", link.download, link.href].join(
        ":",
    );

    const evt = new MouseEvent("click", {
        view: window,
        bubbles: true,
        cancelable: true,
    });

    link.dispatchEvent(evt);
    link.remove();
};

function ChangeEmail() {
    const { user, isLoading } = useUser();

    const [show, setShow] = useState(false);
    const handleClose = () => setShow(false);
    const handleShow = () => setShow(true);

    return (
        <>
            <div className="d-flex justify-content-between placeholder-glow">
                {isLoading || !user ? (
                    <>
                        <code className="placeholder col-3" />
                        <span className="placeholder col-1 " />
                    </>
                ) : (
                    <>
                        <code>{user.email}</code>
                        <span
                            className={styles.verified}
                            data-verified={user.emailValidated}
                        />
                    </>
                )}
            </div>
            <Button variant="outline-dark" onClick={handleShow}>
                Change email
            </Button>

            <Modal
                show={show}
                onHide={handleClose}
                centered
                size="lg"
                fullscreen="md-down"
                backdrop="static"
                keyboard={false}
            >
                <Modal.Header closeButton>
                    <Modal.Title>Change your email address</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <Item
                        title="New email address"
                        description="Please enter your new email address and confirm it.
                            You will receive an email on the new address to confirm the change."
                    >
                        <ChangeEmailForm />
                    </Item>
                </Modal.Body>
            </Modal>
        </>
    );
}

function ChangePassword() {
    const [show, setShow] = useState(false);
    const handleClose = () => setShow(false);
    const handleShow = () => setShow(true);
    return (
        <>
            <Button variant="outline-dark" onClick={handleShow}>
                Change password
            </Button>
            <Modal
                show={show}
                onHide={handleClose}
                centered
                size="lg"
                fullscreen="md-down"
                backdrop="static"
                keyboard={false}
            >
                <Modal.Header closeButton>
                    <Modal.Title>Change your Password</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <Item
                        title="New password"
                        description="Please enter your new password and confirm it."
                    >
                        <ChangePasswordForm />
                    </Item>
                </Modal.Body>
            </Modal>
        </>
    );
}

// unset | valid | invalid
type valid = -1 | 1 | 0;

interface ChangeEmailState {
    emailSend?: boolean;
    error?: string;
    details?: string;
    email: valid;
    confirmEmail: valid;
    password: valid;
}

const initialStateEmail: ChangeEmailState = {
    email: -1,
    password: -1,
    confirmEmail: -1,
};

function ChangeEmailForm() {
    const [state, formAction] = useFormState(changeEmail, initialStateEmail);

    return (
        <form action={formAction} className="mt-3 d-flex gap-2 flex-column">
            <FormError>
                {state.error ? (
                    <>
                        {state.error}
                        {state.details && (
                            <div className="mt-2">{state.details}</div>
                        )}
                    </>
                ) : null}
            </FormError>
            <EmailInput
                name="email"
                label="New email address"
                isValid={state.email == 1}
                isInvalid={state.email == 0}
                disabled={state.emailSend}
            />
            <EmailInput
                name="confirmEmail"
                label="Confirm new email address"
                isValid={state.confirmEmail == 1}
                isInvalid={state.confirmEmail == 0}
                invalidFeedback="Emails do not match!"
                disabled={state.emailSend}
            />
            <PasswordInput
                isValid={state.password == 1}
                isInvalid={state.password == 0}
            />
            <SubmitButton disabled={state.emailSend}>
                {state.emailSend
                    ? "Email send, check your inbox!"
                    : "Change email"}
            </SubmitButton>
        </form>
    );
}

function ChangePasswordForm() {
    const [state, formAction] = useFormState(changePassword, {
        currentPassword: -1,
        newPassword: -1,
        confirmNewPassword: -1,
    });

    return (
        <form action={formAction} className="mt-3 d-flex gap-2 flex-column">
            <FormError>{state.message}</FormError>
            <PasswordInput
                name="currentPassword"
                label="Current password"
                isValid={state.currentPassword == 1}
                isInvalid={state.currentPassword == 0}
                disabled={state.passwordChanged}
                invalidFeedback="Invalid or incorrect password!"
            />
            <PasswordInput
                name="newPassword"
                label="New password"
                isValid={state.newPassword == 1}
                isInvalid={state.newPassword == 0}
                disabled={state.passwordChanged}
            />
            <PasswordInput
                name="confirmNewPassword"
                label="Confirm new password"
                invalidFeedback="Passwords do not match!"
                isValid={state.confirmNewPassword == 1}
                isInvalid={state.confirmNewPassword == 0}
                disabled={state.passwordChanged}
            />
            <SubmitButton disabled={state.passwordChanged}>
                {state.passwordChanged
                    ? "Password changed!"
                    : "Change password"}
            </SubmitButton>
        </form>
    );
}

function LastLogin() {
    const { user, isLoading } = useUser();

    const last_login = useDate(user?.last_login);

    return (
        <>
            <div className="d-flex justify-content-between placeholder-glow">
                {isLoading || !user ? (
                    <>
                        <code className="placeholder col-3" />
                        <code className="placeholder col-1" />
                    </>
                ) : (
                    <>
                        <code>{last_login}</code>
                        <code>{user.last_login_ip}</code>
                    </>
                )}
            </div>
        </>
    );
}

async function changeEmail(
    _previousState: ChangeEmailState,
    formData: FormData,
): Promise<ChangeEmailState> {
    const ret: ChangeEmailState = {
        email: -1,
        password: -1,
        confirmEmail: -1,
    };
    if (formData.has("email")) {
        const email: string = formData.get("email")! as string;
        ret["email"] = validateEmail(email) ? 1 : 0;
        if (formData.has("confirmEmail")) {
            ret["confirmEmail"] =
                formData.get("email") === formData.get("confirmEmail") ? 1 : 0;
        }
    }

    if (formData.has("password")) {
        const password: string = formData.get("password")! as string;
        ret["password"] = validatePW(password) ? 1 : 0;
    }

    //Early return if any validation failed
    if (ret["email"] != 1 || ret["confirmEmail"] != 1 || ret["password"] != 1) {
        return ret;
    }

    const res = await fetch("/email/account/changeEmail", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({
            email: formData.get("email"),
            password: formData.get("password"),
        }),
    });

    if (!res.ok) {
        try {
            const data = await res.json();
            return {
                ...ret,
                ...data,
            };
        } catch (e) {
            console.error(e);
            return {
                ...ret,
                error: "An unknown error occurred, please try again later!",
            };
        }
    }
    const data = await res.json();

    return {
        ...ret,
        emailSend: data.success,
        error: undefined,
        details: undefined,
    };
}

async function changePassword(
    _previousState: ChangePasswordState,
    formData: FormData,
): Promise<ChangePasswordState> {
    const ret = {
        currentPassword: -1,
        newPassword: -1,
        confirmNewPassword: -1,
    } as ChangePasswordState;

    if (formData.has("currentPassword")) {
        const currentPassword: string = formData.get(
            "currentPassword",
        )! as string;
        ret.currentPassword = validatePW(currentPassword) ? 1 : 0;
    }
    if (formData.has("newPassword")) {
        const newPassword: string = formData.get("newPassword")! as string;
        ret.newPassword = validatePW(newPassword) ? 1 : 0;
        if (formData.has("confirmNewPassword")) {
            ret.confirmNewPassword =
                formData.get("newPassword") ===
                formData.get("confirmNewPassword")
                    ? 1
                    : 0;
        }
    }

    if (
        ret.currentPassword != 1 ||
        ret.newPassword != 1 ||
        ret.confirmNewPassword != 1
    ) {
        return ret;
    }
    return await changePasswordAction(ret, formData);
}
