import { wrapperLoggedInUser } from "lib/route_wrapper/wrapperLoggedIn";
import { RequestProps } from "lib/route_wrapper/wrapperSession";
import { NextRequest, NextResponse } from "next/server";

import service from "@snip/database";
import { NotFoundError } from "@snip/database/errors";
import {
    GroupData,
    GroupDataWithMembers,
    UserConfigData,
    UserData,
} from "@snip/database/types";

export const GET = wrapperLoggedInUser(GetUserData);

export type UserRouteResponseReturn = {
    user: UserData;
    userConfig?: UserConfigData;
    groups?: GroupData[] | GroupDataWithMembers[];
};

async function GetUserData(
    req: NextRequest,
    { session }: DeepPartialExcept<RequestProps, "session.user">,
) {
    // Might have an optional parameter for config and groups
    const { searchParams } = new URL(req.url);
    const config = searchParams.get("config") === "true";
    const groups = searchParams.get("groups") === "true";
    const groupMembers = searchParams.get("members") === "true";

    if (!groups && groupMembers) {
        return NextResponse.json(
            { error: "Cannot get group members without groups" },
            { status: 400 },
        );
    }

    //Create promises we resolve together at the end
    const userDataP = service.user.getById(session.user.id);
    let userConfigDataP: Promise<UserConfigData> | undefined;
    if (config) {
        userConfigDataP = service.userConfig
            .getByUserId(session.user.id)
            .catch((e) => {
                // Sometimes the user config may not be initialized
                if (e instanceof NotFoundError) {
                    return service.userConfig.insert({
                        user_id: session.user.id,
                        text_font: "Arial",
                    });
                }
                throw e;
            });
    }
    let groupsDataP: Promise<GroupData[] | GroupDataWithMembers[]> | undefined;
    if (groups) {
        groupsDataP = service.group
            .getByUserId(session.user.id)
            .then(async (groups) => {
                if (groupMembers) {
                    const groupMembersPromises = groups.map((group) => {
                        return service.group.getMembers(group.id);
                    });
                    const groupMembersResults =
                        await Promise.all(groupMembersPromises);
                    (groups as GroupDataWithMembers[]).forEach(
                        (group, index) => {
                            group.members = groupMembersResults[index]!;
                            // Add self flag
                            for (let j = 0; j < group.members.length; j++) {
                                if (
                                    group.members[j]!.user_id ===
                                    session.user.id
                                ) {
                                    group.members[j]!.self = true;
                                    break;
                                }
                            }
                        },
                    );
                }
                return groups;
            });
    }

    // Resolve promises
    const [userData, userConfigData, groupsData] = await Promise.all([
        userDataP,
        userConfigDataP,
        groupsDataP,
    ]);

    const ret: UserRouteResponseReturn = {
        user: userData,
        userConfig: userConfigData,
        groups: groupsData,
    };

    return NextResponse.json<UserRouteResponseReturn>(ret, { status: 200 });
}
