import mdx from '@next/mdx';
import bundleAnalyzer from "@next/bundle-analyzer";

const withBundleAnalyzer = bundleAnalyzer({
    enabled: process.env.ANALYZE === "true",
});
const withMDX = mdx();



const filter_client = (resource, context) => {
    switch (resource) {
        case "sharp":
        case "canvas":
        case "mariadb":
        case "@snip/database":
        case "@snip/config/server":
            console.log("You are importing a dangerous module in (client-side)! Ignoring " + resource);
            return true;
        default:
            return false;
    }
};



function patchWasmModuleImport(config, isServer) {
    config.experiments = Object.assign(config.experiments || {}, {
        asyncWebAssembly: true,
        syncWebAssembly: true,
    });
    config.module.rules.push({
        test: /\.wasm$/,
        type: 'webassembly/async',
    });


    // see track https://github.com/vercel/next.js/issues/25852
    if (isServer) {
        config.output.webassemblyModuleFilename = './../static/wasm/[modulehash].wasm';
    } else {
        config.output.environment = { ...config.output.environment, asyncFunction: true };
        config.output.webassemblyModuleFilename = 'static/wasm/[modulehash].wasm';
    }
}



export default (phase, { defaultConfig }) => {
    /**
     * @type {import('next').NextConfig}
     */
    let nextConfig = {
        distDir: "build",
        poweredByHeader: false,
        eslint: {
            dirs: [
                "lib",
                "components",
                "app",
            ],
        },
        reactStrictMode: true,
        pageExtensions: ['js', 'jsx', 'mdx', 'ts', 'tsx'],
        webpack: (config, { isServer, webpack }) => {
            // Enable wasm loading
            patchWasmModuleImport(config, isServer);

            if (!isServer) {
                // Ignore fs, path, ... on the client
                config.plugins.push(new webpack.IgnorePlugin({ checkResource: filter_client }));
                // Remove 'node:' from import specifiers, because Next.js does not yet support node: scheme
                // https://github.com/vercel/next.js/issues/28774
                config.plugins.push(
                    new webpack.NormalModuleReplacementPlugin(
                        /^node:/,
                        (resource) => {
                            resource.request = resource.request.replace(/^node:/, '');
                        },
                    ),
                );

                config.resolve.fallback = {
                    "fs": false,
                    "os": false,
                    "path": false,
                };
            }

            //config.externals.push({ "@snip/database": '@snip/database' });
            //config.externals.push({ "@snip/render": '@snip/render' });
            return config;
        },
        experimental: {
            outputFileTracingExcludes: {
                "*": [
                    "node_modules/@swc/core-linux-x64-gnu",
                    "node_modules/@swc/core-linux-x64-musl",
                    "node_modules/@esbuild/linux-x64",
                ],
            },
        }
    }


    // Remove all console logs in production
    if (process.env.NODE_ENV === "production") {
        nextConfig = {
            ...nextConfig,
            compress: true,
            compiler: {
                removeConsole: {
                    exclude: ['error'],
                },
            }
        }
    }


    return withBundleAnalyzer(withMDX(nextConfig));
}