import request from "supertest";
import { server } from "../server";
import { config } from "@snip/config/server";
import { mockConfig } from "../../test_setup/mockConfig";

describe("/render (json)", () => {
    beforeAll(() => {
        server.listen(3000);
    });

    afterAll(() => {
        server.close();
    });

    it("should return 400 if no search params are provided", async () => {
        await request(server).get("/render").expect(400);
    });

    it("should return 400 if invalid search params are provided", async () => {
        await request(server)
            .get("/render")
            .send({ page_id: "invalid" })
            .expect(400);

        await request(server)
            .get("/render")
            .send({ page_id: 1, width: "invalid" })
            .expect(400);

        await request(server)
            .get("/render")
            .send({ page_id: 1, width: 200, format: "invalid" })
            .expect(400);

        await request(server)
            .get("/render")
            .send({
                page_id: 1,
                width: 200,
                format: "jpeg",
                encoding: "invalid",
            })
            .expect(400);

        await request(server)
            .get("/render")
            .send({
                page_id: 1,
                width: 200,
                format: "jpeg",
                encoding: "gzip",
                return_stream: "invalid",
            })
            .expect(400);

        await request(server)
            .get("/render")
            .send({
                page_id: 1,
                width: 200,
                format: "jpeg",
                encoding: "gzip",
                return_stream: "true",
                force_render: "invalid",
            })
            .expect(400);
    });

    it("should return 401 if no session is provided", async () => {
        mockConfig.force_api_session = false;
        mockConfig.session = false;
        await request(server)
            .get("/render")
            .send({ page_id: 1, width: 200, format: "jpeg" })
            .expect(401);
        mockConfig.force_api_session = false;
        mockConfig.session = true;
    });

    it("should return 200 if valid", async () => {
        await request(server)
            .get("/render")
            .send({ page_id: 1, width: 200, format: "jpeg" })
            .expect(200)
            .expect("Content-Type", "image/jpeg")
            .expect("Content-Encoding", "gzip");

        await request(server)
            .get("/render")
            .send({ page_id: 1, width: 1400, format: "pdf", encoding: "none" })
            .expect("Content-Type", "application/pdf")
            .expect(200);
    });

    it("should return 200 if send via query", async () => {
        await request(server)
            .get("/render")
            .query({ page_id: 1, width: 200, format: "jpeg" })
            .expect(200)
            .expect("Content-Type", "image/jpeg")
            .expect("Content-Encoding", "gzip");

        await request(server)
            .get("/render")
            .query({ page_id: 1, width: 200, format: "jpeg", encoding: "none" })
            .expect("Content-Type", "image/jpeg")
            .expect(200);
    });
});

describe("/render/:page_id.(png|jpeg|webp)", () => {
    it("should return 404 not found if not an id", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        for (const ending of endings) {
            const res = await request(server)
                .get("/render/9999." + ending)
                .expect(404)
                .expect("Page 9999 not found");
        }
    });

    it("should allow for 'w' and 'width' query params", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        const id = 1;
        for (const ending of endings) {
            await request(server)
                .get(`/render/${id}.${ending}?w=12`)
                .expect(200);

            await request(server)
                .get(`/render/${id}.${ending}?width=12`)
                .expect(200);
        }
    });
});

describe("any /", () => {
    it("should return 404 if wrong endpoint", async () => {
        await request(server).get("/").expect(404);
        await request(server).get("/wrong").expect(404);
    });
});

describe("/rerender/:page_id", () => {
    it("should return 200 on simple request", async () => {
        mockConfig.force_api_session = true;
        await request(server).post("/rerender/1").expect(200);
        mockConfig.force_api_session = false;
    });
});

describe("any /render/book_preview/:book_id", () => {
    it("should return 200 on simple request", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        for (const ending of endings) {
            await request(server)
                .get(`/render/book_preview/1.${ending}`)
                .expect(200);
        }
    });
});
