import service from "@snip/database";

import {
    accessDenied,
    ParsedPermissionACLData,
    perms_by_bearer_token,
    perms_by_ui_token,
} from "./book_permissions";

/**
 * Get the users permission for a given page
 * this function can e.g. be used to check if a user
 * has read access to a specific page without needing the book
 * id.
 *
 * Optionally you can pass a token to verify the user with. An ui token
 * will only give the user read access, while an api token will give the
 * user read, write access.
 *
 * @param page_id - The ID of the book.
 * @param user_id - The ID of the user.
 * @param ui_token_id - Optional token id for authentication using ui tokens.
 * @param bearer_token - Optional token for authentication Only used for
 * api tokens.
 * @returns The permissions of the user for the book.
 */
export async function getPermsPage(
    page_id: number,
    {
        user_id,
        ui_token_id,
        bearer_token,
    }: {
        user_id?: number;
        ui_token_id?: number;
        bearer_token?: string;
    },
): Promise<ParsedPermissionACLData> {
    let perms: ParsedPermissionACLData;

    // Check i user, ui token or api token is used
    if (bearer_token) {
        perms = await perms_by_bearer_token_page(page_id, bearer_token);
    } else if (ui_token_id) {
        perms = await perms_by_ui_token_page(page_id, ui_token_id);
    } else if (user_id) {
        perms = await perms_by_user_page(page_id, user_id);
    } else {
        perms = accessDenied(-1, "unknown");
    }
    return perms;
}

async function perms_by_user_page(
    page_id: number,
    user_id: number,
): Promise<ParsedPermissionACLData> {
    const perms = await service.permission
        .getByUserAndPageIdReduced(user_id, page_id)
        .catch(() => {
            return accessDenied(
                -1,
                "user",
                "User is not allowed to access this book",
            );
        });

    return {
        ...perms,
        auth_method: "user",
    };
}

async function perms_by_ui_token_page(
    page_id: number,
    ui_token_id: number,
): Promise<ParsedPermissionACLData> {
    try {
        const book_id = await service.page
            .getById(page_id)
            .then((page) => page.book_id);
        return perms_by_ui_token(book_id, ui_token_id);
    } catch (e) {
        return accessDenied(-1, "ui_token");
    }
}

async function perms_by_bearer_token_page(
    page_id: number,
    bearer_token: string,
): Promise<ParsedPermissionACLData> {
    try {
        const book_id = await service.page
            .getById(page_id)
            .then((page) => page.book_id);
        return perms_by_bearer_token(book_id, bearer_token);
    } catch (e) {
        return accessDenied(-1, "bearer");
    }
}
