/* This file loads environment variables from .env file
 * and makes them available to the rest of the app. Please
 * use the .env file to configure your app.
 */

import { z } from "zod";

import { isBrowser } from "@snip/common/environment";

// Check if running in browser
if (isBrowser() && process.env.NODE_ENV !== "test") {
    throw new Error(
        "This file should not be loaded in the browser. SECURITY RISK!",
    );
}

const configSchema = z.object({
    NODE_ENV: z.enum(["development", "production", "test"]),
    PORT: z.number(),
    HOSTNAME: z.string(),
    MAINTENANCE: z.boolean(),
    MYSQL: z.object({
        HOST: z.string(),
        PORT: z.number(),
        USER: z.string(),
        PASSWORD: z.string(),
    }),
    EMAIL: z.object({
        HOST: z.string(),
        PORT: z.number(),
        USER: z.string(),
        PASSWORD: z.string(),
        TOKEN_SECRET: z.string().min(32),
        BCC: z.string().email().optional(),
        FROM: z.string().email().optional(),
    }),
    SOCKET: z.object({
        NUM_PROCESSES: z.number(),
        SECRET: z.string(),
    }),
    API_TOKEN_SECRET: z.string().min(32),
    SESSION_SECRET: z.string().min(32),
    RENDER: z.object({
        NUM_PROCESSES: z.number(),
        DIR: z.string(),
        FORMATS: z.array(
            z.object({
                format: z.string(),
                widths: z.array(z.number()),
            }),
        ),
    }),
    PACKAGE_JSON: z.string(),
});

const config_ = {
    /** Is set by the startup of the server */
    NODE_ENV: process.env.NODE_ENV || "development",
    /** The port the server is running on */
    PORT: process.env.PORT ? parseInt(process.env.PORT) : 4000,
    /** The host the server is running on */
    HOSTNAME: process.env.HOSTNAME || "localhost",

    MAINTENANCE: process.env.MAINTENANCE === "true",

    /** Everything database related i.e. mariadb/mysql */
    MYSQL: {
        /** The host of the mysql server */
        HOST: process.env.MYSQL_HOST || "localhost",
        /** The port of the mysql server */
        PORT: process.env.MYSQL_PORT ? parseInt(process.env.MYSQL_PORT) : 3306,
        /** The user of the mysql server */
        USER: process.env.MYSQL_USER || "snip",
        /** The password of the mysql server */
        PASSWORD:
            process.env.MYSQL_PASSWORD || "well_this_should_be_changed_as_well",
    },

    /** Everything regarding email verification */
    EMAIL: {
        /** The host of the email server */
        HOST: process.env.EMAIL_HOST || "email.gwdg.de",
        /** The port of the email server */
        PORT: process.env.EMAIL_PORT ? parseInt(process.env.EMAIL_PORT) : 587,
        /** The user of the email server */
        USER: process.env.EMAIL_USER || "",
        /** The password of the email server */
        PASSWORD: process.env.EMAIL_PASSWORD || "",
        /** The secret key which is used for the tokens */
        TOKEN_SECRET:
            process.env.EMAIL_TOKEN_SECRET ||
            "keyboard_cat_at_least_32_chars_long",

        /** The email address where an copy of each email is sent
         */
        BCC: process.env.EMAIL_BCC,

        /** The email address where the email is sent from */
        FROM: process.env.EMAIL_FROM,
    },

    /** Everything related to the socket connection */
    SOCKET: {
        /** The number of socket processes */
        NUM_PROCESSES: process.env.SOCKET_NUM_PROCESSES
            ? parseInt(process.env.SOCKET_NUM_PROCESSES)
            : 2,
        /** Secret used for initial verification of socket connection (deprecated)*/
        SECRET:
            process.env.SOCKET_SECRET || "keyboard_cat_at_least_32_chars_long",
    },

    /** Secret used to hash the api tokens */
    API_TOKEN_SECRET:
        process.env.API_TOKEN_SECRET || "keyboard_cat_at_least_32_chars_long",

    /** Secret used to encryption/decryption session cookies */
    SESSION_SECRET:
        process.env.SESSION_SECRET || "keyboard_cat_at_least_32_chars_long",

    RENDER: {
        /** The number of render processes */
        NUM_PROCESSES: process.env.NUM_RENDER_PROCESSES
            ? parseInt(process.env.NUM_RENDER_PROCESSES)
            : 4,

        /** The folder for image storage */
        DIR: process.env.RENDER_DIR || "/usr/src/app/images",

        /** Targets for the render worker, these are
         * the files and sizes that can be requested
         *
         * more targets might slow down rerendering
         */
        FORMATS: [
            {
                format: "jpeg",
                widths: [150, 720],
            },
            {
                format: "pdf",
                widths: [720],
            },
        ],
    },

    /** Package json location for version */
    PACKAGE_JSON: process.env.PACKAGE_JSON || "package.json",
};

// Validate the config
const parse = configSchema.safeParse(config_);

if (parse.success === false) {
    throw new Error(parse.error.message);
    process.exit(1);
}

export const config = parse.data as typeof config_;
