import { BaseData, BaseView } from "@snip/database/types";

import { BaseSnip, BaseSnipArgs, RenderContext, SnipData } from "./base";

export interface ArrayData extends BaseData {
    snips: SnipData<BaseData, BaseView>[];
    legacy: boolean;
}

export interface ArrayView extends BaseView {
    snip_order: number[] | null;
}

export interface ArraySnipArgs extends BaseSnipArgs {
    /** The underlying snips of the array snip
     *
     * default: []
     */
    snips?: BaseSnip[];

    /** Whether the array snip is a legacy array snip
     *
     * This defines how the array snip is rendered
     * if legacy is true, the snips are ordered
     * vertically, otherwise they aren't
     *
     * default: false
     */
    legacy?: boolean;

    /** The order of the snips in the array snip
     *  This defines how the snips are ordered.
     *
     * This might be important if one snip should be
     * rendered on top of another snip.
     *
     * E.g. [2, 0, 1] would order the snips in the
     * array snip as snips[2], snips[0], snips[1]
     *
     * default: null
     */
    snip_order?: number[];
}

export class ArraySnip extends BaseSnip {
    public type = "array";

    public snips: BaseSnip[];
    public legacy: boolean;
    public snip_order: number[];

    // Indicate if all snip images are loaded
    ready: Promise<boolean>;

    constructor({
        snips = [],
        legacy = false,
        snip_order,
        ...baseArgs
    }: ArraySnipArgs) {
        super({ ...baseArgs });
        this.snips = snips;
        this.legacy = legacy;

        // Set the snip order
        if (snip_order) {
            this.snip_order = snip_order;
        } else {
            this.snip_order = [...Array(this.snips.length).keys()];
        }

        this.ready = this.checkReady();
    }

    /** Check all snips loaded
     * e.g. images prepared
     */
    async checkReady(): Promise<boolean> {
        const promises: Promise<boolean>[] = [];
        for (const snip of this.snips) {
            if (Object.hasOwn(snip, "ready")) {
                promises.push((snip as ArraySnip).ready);
            }
        }
        await Promise.all(promises);
        return true;
    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<ArrayData, ArrayView>} data - Data dict
     */
    static from_data(
        data: SnipData<ArrayData, ArrayView>,
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        get_snip_from_data?: any,
    ): ArraySnip {
        // Get snips from data
        const snips = data.data?.snips.map((snip_data) => {
            return get_snip_from_data(snip_data);
        });
        return new ArraySnip({
            //Array
            snips,
            legacy: data.data.legacy,
            snip_order: data.view?.snip_order || undefined,
            // Base
            x: data.view?.x,
            y: data.view?.y,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    /** Render an array snip
     *
     * This also renders the underlying snips these should not
     * be rendered separately!
     *
     * @param {RenderContext} ctx - The canvas context
     */
    render(ctx: RenderContext): void {
        const transform = ctx.getTransform();
        // Draw snips
        ctx.translate(this._x, this._y);
        if (this.legacy) {
            this.snip_order.forEach((index) => {
                this.snips[index]!.render(ctx);
                ctx.translate(0, this.snips[index]!.height);
            });
        } else {
            this.snip_order.forEach((index) => {
                this.snips[index]!.render(ctx);
            });
        }

        // Restore transform
        ctx.setTransform(transform);
    }

    to_data(minimal?: boolean): SnipData<ArrayData, ArrayView> {
        const data = super.to_data(minimal) as SnipData<ArrayData, ArrayView>;
        data.data.snips = [];
        for (const snip of this.snips) {
            data.data.snips.push(snip.to_data(minimal));
        }
        data.data.legacy = this.legacy;

        data.view.snip_order = this.snip_order;
        return data;
    }

    get height(): number {
        if (this.legacy) {
            let height = 0;
            for (const snip of this.snips) {
                height += snip.height + Math.max(snip.y, 0);
            }
            return height;
        }
        let y_min = this._y;
        let y_max = this._y;
        for (const snip of this.snips) {
            y_min = Math.min(y_min, this._y + snip.y);
            y_max = Math.max(y_max, this._y + snip.y + snip.height);
        }
        return Math.abs(y_max - y_min);
    }
    get width(): number {
        let x_min = this._x;
        let x_max = 0;
        for (const snip of this.snips) {
            x_min = Math.min(x_min, this._x + snip.x);
            x_max = Math.max(x_max, this._x + snip.x + snip.width);
        }
        return x_max - x_min;
    }

    /** Returns the offset of the snip if the plotting type is legacy
     *
     */
    offset_snip(snip_id: number | null): number {
        if (this.legacy && snip_id !== null) {
            let offset = 0;
            for (const i of this.snip_order) {
                if (this.snips[i]!.id === snip_id) {
                    return offset;
                }
                offset += this.snips[this.snip_order[i]!]!.height;
            }
        }
        return 0;
    }

    /** Setter for page id
     * allows for nested snips to be placed on
     * page.
     */
    public set page_id(value: number | undefined) {
        this._page_id = value;
        for (const snip of this.snips) {
            snip.page_id = value;
        }
    }

    public get page_id(): number | undefined {
        return this._page_id;
    }

    public async place_on_page(page_id: number): Promise<void> {
        this.page_id = page_id;

        for (const snip of this.snips) {
            snip.socket = this.socket;
            await snip.upsert();
        }
        await super.place_on_page(page_id);
    }
}
