import { BaseData, BaseSnip, BaseSnipArgs, BaseView, SnipData } from "./base";

export interface PlaceholderData extends BaseData {
    originalType: string;
}

export interface PlaceholderView extends BaseView {
    width: number;
    height: number;
}

export interface PlaceholderSnipArgs extends BaseSnipArgs {
    originalType: string;
    width: number;
    height: number;
}

/** A placeholder snip can be used
 * in places where it might be too expensive to transfer a full
 * snip. This was mainly introduced to streamline the event
 * protocol.
 *
 * For instance instead of transferring image data we just transfer
 * a snip with the height, width precomputed.
 *
 */
export class PlaceholderSnip extends BaseSnip {
    type: string = "placeholder";

    height: number;
    width: number;
    originalType: string;

    constructor({
        originalType,
        height,
        width,
        ...baseArgs
    }: PlaceholderSnipArgs) {
        super({ ...baseArgs });
        this.width = width;
        this.height = height;
        this.originalType = originalType;
    }

    static from_data(
        data: SnipData<PlaceholderData, PlaceholderView>,
    ): PlaceholderSnip {
        return new PlaceholderSnip({
            // placeholder
            originalType: data.data.originalType,
            height: data.view?.height,
            width: data.view?.width,
            // Base
            x: data.view?.x,
            y: data.view?.y,
            rot: data.view?.rot,
            mirror: data.view?.mirror,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    static from_snip(snip: BaseSnip): PlaceholderSnip {
        const data: SnipData<PlaceholderData, PlaceholderView> = {
            view: {
                x: snip.x,
                y: snip.y,
                width: snip.width,
                height: snip.height,
                rot: snip.rot,
                mirror: snip.mirror,
            },
            data: {
                originalType: snip.type,
                type: PlaceholderSnip.prototype.type,
            },
            id: snip.id,
            page_id: snip.page_id,
            book_id: snip.book_id,
            last_updated: snip.last_updated,
            created: snip.created,
        };

        return PlaceholderSnip.from_data(data);
    }

    to_data(): SnipData<PlaceholderData, PlaceholderView> {
        const data = super.to_data();
        return {
            ...data,
            view: {
                ...data.view,
                x: this.x,
                y: this.y,
                width: this.width,
                height: this.height,
                rot: this.rot,
                mirror: this.mirror,
            },
            data: {
                ...data.data,
                originalType: this.originalType,
                type: this.type,
            },
        };
    }

    public render(): void {
        throw new Error("Method not implemented.");
    }
}
