import { isBrowser, isWorker } from "@snip/common/environment";
import { AllowedFonts, FONT_URLS } from "@snip/database/fonts/internal";
import { BaseData, BaseView } from "@snip/database/types";

import { BaseSnip, BaseSnipArgs, RenderContext, SnipData } from "./base";
import { TextSnipLegacy } from "./legacy/text";

export interface TextData extends BaseData {
    text: string;
    version: number;
}

export interface TextView extends BaseView {
    size: number;
    lheight: number;
    font: string;
    wrap: number;
    colour: string;
    baseline: CanvasTextBaseline;
}

export interface TextSnipArgs extends BaseSnipArgs {
    /** The text to be displayed
     *
     * default: ""
     */
    text?: string;

    /** The font family to use for the text, this
     * must be available on the client and the server
     * for the text to be displayed correctly.
     *
     * default: "PlexMono"
     */
    fontFamily?: string;

    /** The font size of the text
     *
     * See https://gitlab.gwdg.de/irp/snip/-/blob/legacy/public/assets/js/common.js#L1159 for defaults
     *
     * default: 12
     */
    fontSize?: number;

    /** The line height of the text in pixels
     * (i.e. the distance between the baseline of
     * two lines of text)
     *
     * default: fontSize*1.25
     */
    lineHeight?: number;

    /** The colour of the text to be displayed
     *
     * default: "black"
     */
    colour?: string;

    /** The number of characters to wrap the text at
     *
     * This should not be used if possible as it is
     * kinda buggy and not very useful atm.
     *
     * default: 78
     */
    lineWrap?: number;

    /** Version number as used by the database
     *
     * Changelog:
     * 2 - Linewrap is in pixels instead of characters
     */
    version?: number;

    /** Text baseline
     * default: "top"
     */
    baseline?: CanvasTextBaseline;
}

/**
 * Represents a font used in a canvas.
 */
class CanvasFont {
    fontStyle: string;
    fontWeight: string;
    fontFamily: AllowedFonts;
    fontSize: number; //in px
    lineHeight: number; //relative

    /**
     * Creates a new instance of the CanvasFont class.
     * @param options - The font options.
     */
    constructor({
        fontStyle = "normal",
        fontWeight = "normal",
        fontSize = 12,
        lineHeight = 1.25,
        fontFamily = "Arial",
    }: {
        fontStyle?: string;
        fontWeight?: string;
        fontSize?: number;
        lineHeight?: number;
        fontFamily?: string;
    }) {
        this.fontStyle = fontStyle;
        this.fontWeight = fontWeight;
        // Im not sure anymore where the 2.35 comes from but it is needed
        this.fontSize = fontSize * 2.35;
        this.lineHeight = lineHeight;
        if (!FONT_URLS.map((f) => f[1]).includes(fontFamily as AllowedFonts)) {
            throw new Error(`Font family ${fontFamily} is not allowed`);
        }
        this.fontFamily = fontFamily as AllowedFonts;
    }

    /**
     * Gets the font string representation.
     * @returns The font string.
     */
    get fontStr(): string {
        return `${this.fontStyle} ${this.fontWeight} ${this.fontSize}px ${this.fontFamily}`;
    }
}

export class TextSnip extends BaseSnip {
    public type = "text";

    public canvasFont: CanvasFont;
    public baseline: CanvasTextBaseline;
    public colour: string;

    public version: number;

    constructor({
        text = "",
        fontFamily = "PlexMono",
        fontSize = 12,
        colour = "black",
        lineHeight = undefined,
        lineWrap = undefined,
        baseline = "middle",
        ...baseArgs
    }: TextSnipArgs) {
        super({ ...baseArgs });

        // Hardcoded version number
        this.version = 4;

        // Set font
        this.canvasFont = new CanvasFont({
            fontSize,
            fontFamily,
            lineHeight,
        });

        // Set colour
        this.colour = colour;

        this._lineWrap = lineWrap || 400;
        this._raw_text = text;
        this._text = this.formatText_v3(text);

        // Set baseline (should be middle from 1.6 onwards)
        this.baseline = baseline;
    }

    /** Renders the simple text to the canvas function
     *
     * @param {RenderContext} ctx - Canvas context
     */
    render(ctx: RenderContext): void {
        // Set font and style
        // 2480/1400  (TODO put into database update script)
        ctx.font = this.canvasFont.fontStr;
        ctx.fillStyle = this.colour;
        ctx.textAlign = "left";
        ctx.textBaseline = this.baseline;

        // Split text in lines
        // For now only newline is supported
        // linewrap has to be done manually
        const line = this.text.split("\\n");

        // Apply translate, rotate and mirror
        // Base transform in transform function
        // see base.ts
        const transform = ctx.getTransform();
        this.transform(ctx);

        let y = 0;
        if (this.baseline === "middle") {
            // offset to keep the text inside the snip
            // x and y coords
            y += (this.canvasFont.lineHeight * this.canvasFont.fontSize) / 2;
        }

        // Draw text
        line.forEach((text: string) => {
            ctx.fillText(text, 0, y);
            y += this.canvasFont.lineHeight * this.canvasFont.fontSize;
        });

        ctx.setTransform(transform);
    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<TextData, TextView>} data - Data dict
     */
    static from_data(
        data: SnipData<TextData, TextView>,
    ): TextSnip | TextSnipLegacy {
        if (data.data.version < 3) {
            return new TextSnipLegacy({
                // Text
                text: data.data.text,
                fontFamily: data.view.font,
                fontSize: data.view.size,
                colour: data.view.colour,
                lineHeight: data.view.lheight,
                lineWrap: data.view.wrap,
                version: data.data.version,
                //Base
                x: data.view?.x,
                y: data.view?.y,
                rot: data.view?.rot,
                mirror: data.view?.mirror,

                id: data.id,
                page_id: data.page_id,
                book_id: data.book_id,
                last_updated: data.last_updated,
                created: data.created,
            });
        }

        return new TextSnip({
            // Text
            text: data.data.text,
            fontFamily: data.view.font,
            fontSize: data.view.size,
            colour: data.view.colour,
            lineHeight: data.view.lheight,
            lineWrap: data.view.wrap,
            version: data.data.version,
            baseline: data.view.baseline,
            //Base
            x: data.view?.x,
            y: data.view?.y,
            rot: data.view?.rot,
            mirror: data.view?.mirror,

            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    /** Create a json string for mysql from a text snip
     *
     * @returns {Dictionary}
     */
    data(): TextData {
        return {
            type: "text",
            text: this._raw_text,
            version: this.version,
        };
    }
    view(): TextView {
        const view = super.view() as TextView;
        view.font = this.canvasFont.fontFamily;
        view.size = this.canvasFont.fontSize / 2.35;
        view.lheight = this.canvasFont.lineHeight;
        view.colour = this.colour;
        view.wrap = this.lineWrap;
        view.baseline = this.baseline;
        return view;
    }

    public getNumLines(): number {
        return this.text.split("\\n").length;
    }

    get height(): number {
        return (
            this.canvasFont.lineHeight *
            this.canvasFont.fontSize *
            this.getNumLines()
        );
    }
    get width(): number {
        const canvas = this.createCanvas(1, 1);
        const ctx = canvas.getContext(
            "2d",
        ) as OffscreenCanvasRenderingContext2D;
        ctx.font = this.canvasFont.fontStr;
        let maxWidth = 0;
        const lines = this.text.split("\\n");
        for (let i = 0; i < lines.length; i++) {
            maxWidth = Math.max(maxWidth, ctx.measureText(lines[i]!).width);
        }
        return maxWidth;
    }

    private _lineWrap: number;
    get lineWrap(): number {
        return this._lineWrap;
    }
    set lineWrap(lineWrap: number) {
        this._lineWrap = lineWrap;
        if (this.version < 4) {
            this._text = this.formatText_v3(this._raw_text);
        } else {
            this._text = this.formatText(this._raw_text);
        }
    }

    private _text: string;
    private _raw_text: string;
    get text(): string {
        return this._text;
    }
    set text(text: string) {
        this._raw_text = text;
        if (this.version < 4) {
            this._text = this.formatText_v3(text);
        } else {
            this._text = this.formatText(this._raw_text);
        }
    }

    private formatText(inputString: string): string {
        // Replace tabs with 8 spaces
        inputString = inputString.replaceAll(/\t/g, "        ");
        // Replace newlines with \n
        inputString = inputString.replaceAll(/\n/g, "\\n");

        // split text in (presplit)lines
        const lines = inputString.split("\\n");

        // Measure text width for each line and get
        // wrap points
        const font = this.canvasFont.fontStr;
        const output: string[] = [];
        for (let i = 0; i < lines.length; i++) {
            const subsStrings = this.splitStringAtWrap(
                lines[i]!,
                this.lineWrap,
                font,
            );
            output.push(subsStrings.join("\\n"));
        }
        return output.join("\\n");
    }

    /**
     * Splits a string into segments based on a maximum width constraint.
     *
     * This method measures the width of the string as it would be rendered
     * using the specified font. It then splits the string into segments where
     * each segment's width does not exceed the given maximum width.
     *
     * @param str - The input string to be split.
     * @param maxWidth - The maximum width in pixels for each segment.
     * @param font - The font string used for measuring the text width.
     * @returns An array of string segments, each fitting within the specified width.
     */
    splitStringAtWrap(str: string, maxWidth: number, font: string): string[] {
        if (maxWidth == -1) {
            return [str];
        }
        if (!isBrowser() && !isWorker()) {
            return [str];
        }
        const canvas = this.createCanvas(1, 1);
        const ctx = canvas.getContext(
            "2d",
        ) as OffscreenCanvasRenderingContext2D;
        ctx.font = font;

        let width = 0;
        const splits = [];
        for (let c = 0; c < str.length; c++) {
            const char = str.charAt(c);
            const cWidth = ctx.measureText(char).width;
            width += cWidth;
            if (width >= maxWidth) {
                splits.push(c);
                width = cWidth;
            }
        }

        // split into segments
        const segments: string[] = [];
        let start = 0;

        for (const split of splits) {
            segments.push(str.slice(start, split));
            start = split;
        }

        // Add the last segment if there's any remaining part of the string
        if (start < str.length) {
            segments.push(str.slice(start));
        }
        return segments;
    }

    /** Formats the text to a usable format
     * - Replaces tabs with 8 spaces
     * - Replaces newlines with \n
     * - Calculate width and further split text if
     *  lineWrap is set and text width is too long
     *  @deprecated
     */
    formatText_v3(inputString: string): string {
        // Replace tabs with 8 spaces
        inputString = inputString.replaceAll(/\t/g, "        ");
        // Replace newlines with \n
        inputString = inputString.replaceAll(/\n/g, "\\n");

        // split text in (presplit)lines
        const lines = inputString.split("\\n");

        // Measure text width
        const font = this.canvasFont.fontStr;
        for (let i = 0; i < lines.length; i++) {
            const o = this.get_overlap_index_v3(lines[i]!, this.lineWrap, font);
            if (o !== undefined && o > 0) {
                // Split line and add to next line
                if (lines[i + 1] === undefined) lines[i + 1] = "";
                lines[i + 1] = lines[i]!.substring(o) + lines[i + 1];
                lines[i] = lines[i]!.substring(0, o);
            }
            if (o == 0) {
                // each character on a new line
                lines[i] = lines[i]!.split("").join("\\n");
            }
        }

        // Join lines
        return lines.join("\\n");
    }

    /** Returns index of character which
     * overlaps the lineWrap limit.
     *
     * Returns undefined if no character overlaps
     * @deprecated
     */
    get_overlap_index_v3(str: string, maxWidth: number, font: string) {
        if (maxWidth == -1) {
            return undefined;
        }
        if (!isBrowser() && !isWorker()) {
            return undefined;
        }
        const canvas = this.createCanvas(1, 1);
        const ctx = canvas.getContext(
            "2d",
        ) as OffscreenCanvasRenderingContext2D;
        ctx.font = font;

        let width = 0;
        for (let c = 0; c < str.length; c++) {
            const char = str.charAt(c);
            width += ctx.measureText(char).width;
            if (width > maxWidth) {
                return c;
            }
        }
        return undefined;
    }
}
